from pathlib import Path
from typing import List, Optional, Union

import pyarrow as pa
import pyarrow.dataset as ds

from .io import write_dataset
from .query import Query


class Dataset:
    def __init__(
        self,
        location: Union[str, Path],
        partitioning: Optional[List[str]] = None,
        format: str = "parquet",
        filesystem: Optional[pa.fs.FileSystem] = None,
        batch_size: int = 1_000_000,
    ):
        self._location = location
        self._partitioning = partitioning
        self._format = format
        self._fs = filesystem
        self.batch_size = batch_size

        self._dataset = ds.dataset(
            location,
            format=format,
            filesystem=filesystem,
            partitioning=partitioning,
        )

    @property
    def location(self):
        return self._location

    @property
    def schema(self):
        return self._dataset.schema

    @property
    def nrows(self):
        return self._dataset.count_rows()

    @property
    def dataset(self):
        return self._dataset

    @property
    def query(self):
        """Returns a Query object that can be used to query the dataset"""
        return Query(self._dataset)

    def cast(self, **kwarg):
        """Cast a field in the dataset to another data type"""
        new_schema = self._dataset.schema

        for field, data_type in kwarg.items():
            idx = new_schema.get_field_index(field)
            new_schema = new_schema.set(idx, pa.field(field, data_type))

        if self._dataset.count_rows() > self.batch_size:
            # Dataset is too big to fit in memory, so we need to get it in batches and cast columns in each batch
            for batch in self._dataset.scanner(batch_size=self.batch_size).to_batches():
                table = pa.Table.from_batches([batch])
                new_table = table.cast(new_schema)

                dataset = ds.dataset(
                    [new_table],
                    schema=new_schema,
                )

                write_dataset(
                    dataset,
                    self.location,
                    new_schema,
                    format=self._format,
                    partitioning=self._partitioning,
                    filesystem=self._fs,
                    batch_size=self.batch_size,
                )
        else:
            table = self._dataset.to_table()
            new_table = table.cast(new_schema)

            dataset = ds.dataset([new_table], schema=new_schema)

            write_dataset(
                dataset,
                self.location,
                new_schema,
                format=self._format,
                partitioning=self._partitioning,
                filesystem=self._fs,
                batch_size=self.batch_size,
            )

    def to_pandas(self):
        """Returns the dataset as a Pandas DataFrame.

        Be aware that the dataset might be too large to fit in memory. Consider creating a query first."""
        return self._dataset.to_table().to_pandas()

    def __repr__(self):
        return f"Dataset(location={self._location})"



from typing import List, Generator, Optional

import pyarrow.dataset as ds
import pandas as pd


field = ds.field


class Query:
    def __init__(self, dataset: ds.Dataset):
        self._dataset = dataset
        self._columns: Optional[List[str]] = None
        self._filter = None

    def select(self, columns: List[str]) -> "Query":
        """Select specific columns from the dataset"""
        self._columns = columns
        return self

    def where(self, expression: ds.Expression) -> "Query":
        """Filter the dataset, returns the query object.

        Refer to PyArrow documentation for information on creating expressions:
          https://arrow.apache.org/docs/python/generated/pyarrow.dataset.Expression.html
        """
        if self._filter is not None:
            self._filter = self._filter & expression
        else:
            self._filter = expression
        return self

    def count_rows(self) -> int:
        """Return the number of rows to be returned by the current query"""
        nrows: int = self._dataset.scanner(
            columns=self._columns, filter=self._filter
        ).count_rows()

        return nrows

    def to_pandas(self) -> pd.DataFrame:
        """Fetch the data from the dataset and return it as a Pandas DataFrame"""
        return (
            self._dataset.scanner(columns=self._columns, filter=self._filter)
            .to_table()
            .to_pandas()
        )

    def to_batches(self, batch_size=1_000_000) -> Generator[pd.DataFrame, None, None]:
        """Fetch the data from the dataset as batches and return a generator the yields each
        batch as a Pandas DataFrame"""
        for batch in self._dataset.scanner(
            columns=self._columns, filter=self._filter, batch_size=batch_size
        ).to_batches():
            yield batch.to_pandas()

    def __repr__(self):
        return "Query(columns={}, filter={})".format(self._columns, self._filter)



from itertools import chain
from pathlib import Path
from typing import List, Optional, Union

import pyarrow as pa
import pyarrow.dataset as ds


def write_dataset(
    dataset: ds.dataset,
    location: Union[str, Path],
    schema: pa.Schema,
    format="parquet",
    filesystem: Optional[pa.fs.FileSystem] = None,
    partitioning: Optional[List[str]] = None,
    batch_size: int = 1_000_000,
):
    """Write a dataset to a location, appending if there is an existing dataset. This can be tricky
    as the out-of-memory dataset might be too large to fit in memory. So we need to be careful to
    get the data in batches and only the part that needs to be appended to.
    """
    # Get the out-of-memory dataset to append to
    oom_dataset: Optional[ds.Dataset] = None
    try:
        oom_dataset = ds.dataset(
            location,
            schema=schema,
            format=format,
            filesystem=filesystem,
            partitioning=partitioning,
        )
    except FileNotFoundError:
        pass

    if oom_dataset is not None:
        # We need to get the OOM dataset in batches, so we can append to it
        if partitioning is None:
            batch_filter = None
        else:
            partition_cols = dataset.to_table(columns=partitioning)
            unique_col_vals = [
                ds.field(col).isin(partition_cols.column(col).unique())
                for col in partition_cols.column_names
            ]
            batch_filter = unique_col_vals[0]
            if len(unique_col_vals) > 1:
                for each in unique_col_vals[1:]:
                    batch_filter = batch_filter and each

        if oom_dataset is not None:
            oom_batches = oom_dataset.scanner(
                batch_size=batch_size, filter=batch_filter
            ).to_batches()
        else:
            oom_batches = []

        im_batches = dataset.to_batches()

        ds.write_dataset(
            chain(oom_batches, im_batches),
            location,
            schema=schema,
            filesystem=filesystem,
            partitioning=partitioning,
            format=format,
            existing_data_behavior="overwrite_or_ignore",
        )
    else:
        ds.write_dataset(
            dataset,
            location,
            schema=schema,
            filesystem=filesystem,
            partitioning=partitioning,
            format=format,
            existing_data_behavior="overwrite_or_ignore",
        )



import os
from typing import Dict, List, Optional, Union
from pathlib import Path

import pandas as pd
import pyarrow as pa
import pyarrow.dataset as ds
import pyarrow.fs as fs

from .partitioning import Partitioning
from .io import write_dataset
from .dataset import Dataset


class Stowage:
    def __init__(
        self,
        location: str,
        schemas: Optional[Dict[str, pa.Schema]] = None,
        partitioning: Optional[Union[Partitioning, str]] = None,
        format="parquet",
        access_key: Optional[str] = None,
        secret_key: Optional[str] = None,
        region: Optional[str] = None,
    ):
        if schemas is None:
            self._schemas = {}
        else:
            self._schemas = schemas
        self._format = format
        self._fs: fs.FileSystem = fs.LocalFileSystem()
        self._datasets: Dict[str, Dataset] = {}

        if isinstance(partitioning, str):
            self._partitioning = Partitioning(partitioning)
        elif partitioning is None:
            self._partitioning = Partitioning()
        else:
            self._partitioning = partitioning

        path = location
        if location.startswith("s3://"):
            self._fs = fs.S3FileSystem(
                access_key=os.environ.get("AWS_ACCESS_KEY_ID", access_key),
                secret_key=os.environ.get("AWS_SECRET_ACCESS_KEY", secret_key),
                region=os.environ.get("AWS_REGION", region),
            )
            path = location[5:]
        self._path = Path(path)

    @property
    def schema(self):
        return self._schemas

    @property
    def format(self):
        return self._format

    def list_contents(self):
        files: List[fs.FileInfo] = self._fs.get_file_info(
            fs.FileSelector(str(self._path))
        )
        for file in files:
            print(file.path.split("/")[-1])

    def update_schemas(self, schema: Dict[str, pa.Schema]):
        self._schemas.update(schema)

    def load_dataset(self, name: str) -> Dataset:
        """Load a dataset from the filesystem"""
        dataset = self._datasets[name] = Dataset(
            str(self._path / name),
            format=self._format,
            filesystem=self._fs,
            partitioning=self._partitioning[name],
        )
        self._schemas[name] = dataset.schema
        return self._datasets[name]

    def create_from_dict(self, name: str, data: Dict[str, List]) -> ds.Dataset:
        """Create a dataset from a dictionary"""
        if name not in self._schemas:
            raise ValueError(f"Dataset {name} not found in schema")

        schema = self._schemas[name]

        # If a column exists in the schema, but it's not in the data, fill it with Nones
        data_length = len(data[[each for each in data][0]])
        for field in schema.names:
            if field not in data:
                data[field] = [None] * data_length

        table = pa.Table.from_pydict(data, schema=schema)
        dataset = self._datasets[name] = ds.dataset([table])

        return dataset

    def to_pandas(
        self,
        name: str,
        columns: Optional[List[str]] = None,
        filter: Optional[ds.Expression] = None,
    ) -> pd.DataFrame:
        """Convert a dataset to a pandas DataFrame. Please note that the resulting DataFrame
        might not fit in memory. Request columns or filter the dataset to reduce the size."""
        if name not in self._datasets:
            dataset = self.load_dataset(name)
        else:
            dataset = self._datasets[name]

        return dataset.query.select(columns).where(filter).to_pandas()

    def save_dataset(self, name: str):
        """Saves dataset to location, appends if data is already there"""
        if name not in self._datasets:
            raise ValueError(f"Dataset {name} not found")

        # In-memory dataset
        dataset = self._datasets[name]

        write_dataset(
            dataset,
            str(self._path / name),
            self._schemas[name],
            format=self.format,
            filesystem=self._fs,
            partitioning=self._partitioning[name],
        )

    def __getitem__(self, name: str):
        if name not in self._datasets:
            raise ValueError(f"Dataset {name} not found")
        return self._datasets[name]

    def __delitem__(self, name: str):
        del self._datasets[name]



from typing import List, Dict, Union, Optional


class Partitioning:
    def __init__(self, default: Optional[Union[str, List[str]]] = None):
        self.default: Optional[Union[str, List[str]]] = None
        if default is not None and isinstance(default, str):
            self.default = [default]
        elif default is not None:
            self.default = default

        self._partitions: Dict[str, List[str]] = {}

    @staticmethod
    def from_dict(
        partitions: Dict[str, List[str]], default: Optional[Union[str, List[str]]] = None
    ):
        partitioning = Partitioning(default)
        partitioning.update(partitions)
        return partitioning

    def update(self, value: Dict[str, List[str]]):
        self._partitions.update(value)

    def __getitem__(self, key: str):
        return self._partitions.get(key, self.default)

    def __setitem__(self, key: str, val: List[str]):
        self._partitions[key] = val
